package milab.lab04.PatientAgent;

import jade.core.Agent;
import jade.core.AID;
import jade.core.behaviours.*;
import jade.lang.acl.ACLMessage;
import jade.lang.acl.MessageTemplate;
import jade.domain.DFService;
import jade.domain.FIPAException;
import jade.domain.FIPAAgentManagement.DFAgentDescription;
import jade.domain.FIPAAgentManagement.ServiceDescription;

import javax.swing.JComboBox;
import javax.xml.bind.*;

import java.io.File;
import java.util.List;

import milab.*;
import milab.lab04.PatientAgent.language01.*;
import milab.lab04.DoctorAgent.language01.*;

// Patient JADE agent for Bayes reasoning doctor JADE agents (c) Kovcs Dniel Lszl [dkovacs@mit.bme.hu]
public class PatientAgent extends Agent {
	
	// The list of known DoctorAgent agents' AIDs
	public AID[] doctorAgents;
	// Contains the filepath of the file holding the name of main parameter variable (mpv) and the symptoms 
	public String parametersFilePath;

	// Target parameter variable (the target of the Doctors' inference) - The Main Question! :)
	public String targetID;
	// Contains the names of parameter variables describing the PatientAgent's symptoms
	public String[] symptomIDs;
	
	// Contains the (most actual) probability or "x" of every DoctorAgent for the main parameter variable
	public String[] mpvDiagnoses;

  	// The GUI of the PatientAgent
  	private PatientAgentGui myGui;
  
  	// Put agent initializations here
	protected void setup() {

		// Printout a welcome message
		System.out.println("Hello! Patient-agent " + getAID().getName() + " is ready.");

		// Get the location of the knowledge base and the type of sensors as a start-up argument
		Object[] args = getArguments();

		if (args != null && args.length > 0) {

		    parametersFilePath = (String)args[0];
		    File parametersFile = new File(parametersFilePath);

		    if (parametersFile.isFile()) {

		    	System.out.println("Parameter description file is: " + parametersFilePath);

		    	// Now read the parameter variables' name and the MPVi from the specified file
		    	String[] mpvAndSymptomIDs = DataHandler.readFile2String(parametersFilePath).split("\t");

		    	if (mpvAndSymptomIDs.length > 1) {

		    		targetID = mpvAndSymptomIDs[0];
		    		symptomIDs = new String[mpvAndSymptomIDs.length-1];
		    		
		    		for (int i = 1; i < mpvAndSymptomIDs.length; i++) symptomIDs[i-1] = mpvAndSymptomIDs[i];
		    		System.out.println("Parameter variables' names parsed successfully...");
		    			
		    		// Create and show the GUI 
		    		myGui = new PatientAgentGui(this);
		    		myGui.show();

		    	} else {
			  		
					// Make the agent terminate
				    System.out.println("[PARAMETER FILE ERROR] The parameter description file has to declare the name of the target parameter and at least one symptom. Terminating...");
				    doDelete();

			  	}
		    	
		    } else {
		  		
				// Make the agent terminate
			    System.out.println("[PARAMETER FILE ERROR] The specified parameter description file is non existing. Terminating...");
			    doDelete();

		  	}
		    
		} else {

			// Make the agent terminate
		    System.out.println("[PARAMETER FILE ERROR] No parameter description file specified. Terminating...");
		    doDelete();

		}
	    	  
	}

	// Put agent clean-up operations here
	protected void takeDown() {
  	
	  	// Close the GUI
	  	myGui.dispose();
		
		// Printout a dismissal message
		System.out.println("Patient-agent " + getAID().getName() + " terminating.");
    
	}

	public void activateBehaviour(String b) {
	  	
		if (b.equals("InitializeNewRound")) {
			
			addBehaviour(new InitializeNewRound());
			
		} else if (b.equals("RequestDiagnoses")) {
			
			addBehaviour(new RequestDiagnoses());

		} 
	
	}
	
	/**
	   Inner class InitializeNewRound
	   This is the behaviour used by PatientAgent agents to discover the current DoctorAgents
	   on the platform, and then perform some initializations on the GUI and inner variables.
	 */
	private class InitializeNewRound extends OneShotBehaviour {
		
		public void action() {
			
			// Update the list of doctor agents
			DFAgentDescription template = new DFAgentDescription();
			ServiceDescription sd = new ServiceDescription();
			sd.setType("bayes-diagnostic-service");
			template.addServices(sd);

			try {

				DFAgentDescription[] result = DFService.search(myAgent, template); 
				System.out.println("Found the following doctor agents:");
				doctorAgents = new AID[result.length];

				for (int i = 0; i < result.length; ++i) {

					doctorAgents[i] = result[i].getName();
					System.out.println(doctorAgents[i].getName());

				}

			} catch (FIPAException fe) {
	
				fe.printStackTrace();

			}

			// Declare the space usage of mpvDiagnoses array (it is necessary for later being able 2 give value to it)
			mpvDiagnoses = new String[doctorAgents.length];
			
			// Initialize mpvDiagnoses (doctorsDiagnoses come later in the replies) if there is at least 1 doc.
			for (int i = 0; i < doctorAgents.length; i++) {

		    	mpvDiagnoses[i] = "x";

		    }

			// NOW COMES THE GUI UPDATE!!!
					
			if (doctorAgents.length > 0) {

				for (int i = 0; i < doctorAgents.length; i++) {

					myGui.doctors.addItem(doctorAgents[i].getLocalName());

			    }

				myGui.diagButton.setEnabled(true);
				
				for (int i = 0; i < symptomIDs.length; i++) {

					myGui.symptom_values[i].setEnabled(true);
					myGui.symptom_values[i].setSelectedIndex(2);
					
			    }
				
			} else {
				
				myGui.diagButton.setEnabled(false);
				
			}

		} // End of method InitializeNewRound.action()

	} // End of inner class InitializeNewRound

	/**
	   Inner class RequestDiagnoses.
	   This is the behaviour used by PatientAgent agents to request a diagnosis from every known doctor,
	   and get the replies, and update the GUI accordingly.
	 */
	private class RequestDiagnoses extends Behaviour {

	  private AID bestSeller;			// The agent who provides the best offer 
	  private int bestPrice;			// The best offered price
	  private int counter = 0;			// The counter of request-reply pairs from-to doctor agents
	  private MessageTemplate mt;		// The template to receive replies
	  private int step = 0;

	  public void action() {

	    switch (step) {

		    case 0:
		    	
		    	// Send a REQUEST to the next doctor in the row
		    	if (counter < doctorAgents.length) {

		    		ACLMessage msg = new ACLMessage(ACLMessage.REQUEST);
		    		msg.addReceiver(doctorAgents[counter]);
			    	milab.lab04.PatientAgent.language01.ObjectFactory factory = new milab.lab04.PatientAgent.language01.ObjectFactory();
			    	MessageType msgt = factory.createMessageType();

			    	// Getting the actual symptoms from the GUI
			    	
			    	String symptoms;

		    		if (myGui.symptom_values[0].getSelectedItem().equals("false")) {
		    			
		    			symptoms = "0";
		    			
		    		} else if (myGui.symptom_values[0].getSelectedItem().equals("true")) {
		    			
		    			symptoms = "1";
		    			
		    		} else {
		    			
		    			symptoms = "x";
		    			
		    		}

			    	for (int i = 1; i < symptomIDs.length; i++) {
			    		
			    		if (myGui.symptom_values[i].getSelectedItem().equals("false")) {
			    			
			    			symptoms = symptoms + " 0";
			    			
			    		} else if (myGui.symptom_values[i].getSelectedItem().equals("true")) {
			    			
			    			symptoms = symptoms + " 1";
			    			
			    		} else {
			    			
			    			symptoms = symptoms + " x";
			    			
			    		}
			    		
			    	}
			    	
			    	msgt.setSymptoms(symptoms);
			    	
			    	// There should be at least 1 doctor, because NEXT button was pressed on GUI, which
			    	// means exactly this. It is enabled only when there are at least 1 doctors.
				    String mpvds = mpvDiagnoses[0];
				    for (int i = 1; i < mpvDiagnoses.length; ++i){

				    	mpvds = mpvds + " " + mpvDiagnoses[i];

				    }

			    	msgt.setDiagnoses(mpvds);
			    	JAXBElement<MessageType> messageElement = factory.createMessage(msgt);	// Creating the Java Object corresponding to the complete XML

			    	JAXBMarshaller jaxbMarshaller = new JAXBMarshaller();																					// Creating a universal Marshaller
			    	msg.setContent(jaxbMarshaller.generateXMLString("UTF-8", messageElement, "milab.lab04.PatientAgent.language01"));	// Marshalling the Java Object into a String with appropriate XML syntax

			    	msg.setConversationId("patient-diagnose");
			    	msg.setReplyWith("msg" + System.currentTimeMillis()); // Unique value
			      
			    	myAgent.send(msg);
			    	
			    	// Prepare the template to get appropriate replies
			    	mt = MessageTemplate.and(MessageTemplate.MatchConversationId("patient-diagnose"),
			                               	 MessageTemplate.MatchInReplyTo(msg.getReplyWith()));

			    	step = 1;

		    	} else { // Theoretically this case should never occur...

		    		step = 2;

		    	}

		    	break;

		    case 1:

		    	// Receive the diagnose from the doctor agent in the row
		    	ACLMessage reply = myAgent.receive(mt);

		    	if ((reply != null) && (reply.getPerformative() == ACLMessage.INFORM) && (reply.getSender().equals(doctorAgents[counter]))) {
       	
		    		// This is a diagnosis
		    		JAXBUnMarshaller jaxbUnMarshaller = new JAXBUnMarshaller();		// Creating an UnMarshaller
		    		JAXBElement<String> diagnoseElement = (JAXBElement<String>)jaxbUnMarshaller.readXMLString(reply.getContent(), "/jade/src/milab/lab04/doctoragent/language01/language01.xsd", "milab.lab04.DoctorAgent.language01");	// Taking the ACLMessageContent String as an XML, validating it, and unmarshalling it into an appropriate Java Object.
		    		mpvDiagnoses[counter] = diagnoseElement.getValue();			// Update Dr.X's diagnose

		    		counter++;	// Ok, so there was an appropriate reply at a request...

		    		System.out.println("Diagnosis from Dr. " + reply.getSender().getLocalName() + " (" + counter + "/" + doctorAgents.length + ") received...");

		    		if (counter >= doctorAgents.length) {

		    			step = 2; // We've received all replies
		    			
		    			// REFRESH+UPDATE GUI (AND ENABLE EVERYTHING)
						
						myGui.newButton.setEnabled(true);
						myGui.diagButton.setEnabled(true);
						myGui.doctors.setEnabled(true);
						myGui.doctors.addActionListener(myGui.cbal);

						myGui.diagnose_value.setText(mpvDiagnoses[myGui.doctors.getSelectedIndex()]); // !!!!!

						
						
		    		} else {
		    			
		    			step = 0; // We should send and receive more (on the next activation of this behav.)...
		    			
		    		}

		    	} else {

		    		block();

		    	}

		    	break;

	    } // End of switch (step)

	  } // End of method RequestDiagnoses.action()

	  public boolean done() {

		return (step == 2);

	  } // End of method RequestDiagnoses.done()
	  
	}  // End of inner class RequestDiagnoses

}
